package com.ruoyi.business.config;

import com.alibaba.fastjson.JSONObject;
import com.ruoyi.business.utils.PayUtils;
import com.ruoyi.common.utils.security.Md5Utils;
import com.wechat.pay.contrib.apache.httpclient.WechatPayHttpClientBuilder;
import com.wechat.pay.contrib.apache.httpclient.auth.PrivateKeySigner;
import com.wechat.pay.contrib.apache.httpclient.auth.Verifier;
import com.wechat.pay.contrib.apache.httpclient.auth.WechatPay2Credentials;
import com.wechat.pay.contrib.apache.httpclient.auth.WechatPay2Validator;
import com.wechat.pay.contrib.apache.httpclient.cert.CertificatesManager;
import com.wechat.pay.contrib.apache.httpclient.util.AesUtil;
import com.wechat.pay.contrib.apache.httpclient.util.PemUtil;
import lombok.Data;
import org.apache.http.impl.client.CloseableHttpClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.io.ClassPathResource;

import javax.servlet.http.HttpServletRequest;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.security.PrivateKey;
import java.security.Signature;
import java.time.Duration;
import java.time.Instant;
import java.util.Base64;
import java.util.Objects;
import java.util.stream.Collectors;

import static com.wechat.pay.contrib.apache.httpclient.constant.WechatPayHttpHeaders.*;


@Configuration
@Data //使用set方法将wxpay节点中的值填充到当前类的属性中
public class WxPayConfig {

    private static final Logger log = LoggerFactory.getLogger("WxPayConfig");

    /**
     * 加载商户私钥
     */
    private PrivateKey getPrivateKey(String fail) {
        try {
            //示例：私钥存储在文件
            return PemUtil.loadPrivateKey(new FileInputStream(new ClassPathResource(fail).getFile()));
//            return PemUtil.loadPrivateKey(new FileInputStream(new File("D://tuodan/"+fail)));
        } catch (FileNotFoundException e) {
            throw new RuntimeException("私钥文件不存在", e);
        } catch (IOException e) {
            throw new RuntimeException("私钥文件不存在", e);
        }
    }

//    /**
//     * 加载商户私钥
//     */
//    private PrivateKey getPrivateKeyByStr(String privateKey) {
//        try {
//            //示例：私钥为String字符串
//            return PemUtil.loadPrivateKey(new ByteArrayInputStream(privateKey.getBytes(StandardCharsets.UTF_8)));
//        } catch (Exception e) {
//            throw new RuntimeException("私钥文件不存在", e);
//        }
//    }


    /**
     * 签名验证器
     *
     * @return Verifier
     */
    @Bean
    public Verifier getVerifier() throws Exception {

        //获取商户私钥
        PrivateKey privateKey = getPrivateKey(PayUtils.privateKeyPath);
        //私钥签名对象
        PrivateKeySigner keySigner = new PrivateKeySigner(PayUtils.mchSerialNo, privateKey);
        //身份认证对象
        WechatPay2Credentials pay2Credentials = new WechatPay2Credentials(PayUtils.mchId, keySigner);
        // 获取证书管理器实例
        CertificatesManager certificatesManager = CertificatesManager.getInstance();
        // 向证书管理器增加需要自动更新平台证书的商户信息
        certificatesManager.putMerchant(PayUtils.mchId, pay2Credentials,
                //对称加密的秘钥
                PayUtils.apiV3Key.getBytes(StandardCharsets.UTF_8));
        // ... 若有多个商户号，可继续调用putMerchant添加商户信息

        // 从证书管理器中获取verifier
        Verifier verifier = certificatesManager.getVerifier(PayUtils.mchId);
        return verifier;
    }

    /**
     * 获取http请求对象
     * 每次请求更新平台证书
     *
     * @param verifier
     * @return
     */
    @Bean
    public CloseableHttpClient getWxPayClient(Verifier verifier) {
        //商户私钥
        PrivateKey privateKey = getPrivateKey(PayUtils.privateKeyPath);
        WechatPayHttpClientBuilder builder = WechatPayHttpClientBuilder.create()
                .withMerchant(PayUtils.mchId, PayUtils.mchSerialNo, privateKey)
                .withValidator(new WechatPay2Validator(verifier));
        // ... 接下来，你仍然可以通过builder设置各种参数，来配置你的HttpClient

        // 通过WechatPayHttpClientBuilder构造的HttpClient，会自动的处理签名和验签，并进行证书自动更新
        CloseableHttpClient httpClient = builder.build();
        return httpClient;

    }

    /**
     * 签名校验
     * url https://pay.weixin.qq.com/wiki/doc/apiv3_partner/apis/chapter5_1_13.shtml
     */
    public String verifySign(HttpServletRequest request) throws Exception {
        //检查header
        String[] headers = {WECHAT_PAY_SERIAL, WECHAT_PAY_SIGNATURE, WECHAT_PAY_NONCE, WECHAT_PAY_TIMESTAMP};
        for (String headerName : headers) {
            if (request.getHeader(headerName) == null) {
                log.info("{} is null", headerName);
                return null;
            }
        }
        //检查时间
        String timestamp = request.getHeader(WECHAT_PAY_TIMESTAMP);
        Instant responseTime = Instant.ofEpochSecond(Long.parseLong(timestamp));
        if (Duration.between(responseTime, Instant.now()).abs().toMinutes() >= 5) {
            log.info("超过应答时间");
            return null;
        }
        //获取微信返回的参数
        String data;
        try {
            data = request.getReader().lines().collect(Collectors.joining());
        } catch (IOException e) {
            log.error("获取微信V3回调参数失败",e);
            return null;
        }
        //校验签名
        String nonce = request.getHeader(WECHAT_PAY_NONCE);
        String message =  timestamp + "\n" + nonce + "\n" + data + "\n";
        String serial = request.getHeader(WECHAT_PAY_SERIAL);
        String signature = request.getHeader(WECHAT_PAY_SIGNATURE);
        Verifier verifier = getVerifier();

        if (!verifier.verify(serial, message.getBytes(StandardCharsets.UTF_8), signature)) {
            log.info("签名校验失败");
            return null;
        }
        return data;
    }

    /**
     * v3支付回调数据校验
     */
    public String payRequestDecryption(HttpServletRequest request){
        //校验签名
        String data = null;
        try {
            data = this.verifySign(request);
        } catch (Exception e) {
            e.printStackTrace();
        }
        if (data == null){
            return null;
        }
        JSONObject jsonObject = JSONObject.parseObject(data);
        System.out.println(jsonObject);



        String eventType = jsonObject.getString("event_type");
        String resourceType = jsonObject.getString("resource_type");
        if (!Objects.equals(eventType,"TRANSACTION.SUCCESS") || !Objects.equals(resourceType,"encrypt-resource")){
            log.info("不是支付通知不处理:{}",data);
            return null;
        }
        //参数解密
        JSONObject resource = jsonObject.getJSONObject("resource");
        String ciphertext = resource.getString("ciphertext");
        String nonce = resource.getString("nonce");
        String associatedData = resource.getString("associated_data");
        AesUtil aesUtil = new AesUtil(PayUtils.apiV3Key.getBytes(StandardCharsets.UTF_8));
        String result = null;
        try {
            result = aesUtil.decryptToString(associatedData.getBytes(StandardCharsets.UTF_8),nonce.getBytes(StandardCharsets.UTF_8),ciphertext);
        } catch (GeneralSecurityException e) {
            log.error("微信v3解密异常",e);
        }
        System.out.println("解密之后的数据是："+result);
        return result;
    }


    /**
     * 作用：使用字段appId、timeStamp、nonceStr、package计算得出的签名值
     * 场景：根据微信统一下单接口返回的 prepay_id 生成调启支付所需的签名值
     * @param appId
     * @param timestamp
     * @param nonceStr
     * @param pack package
     * @return
     * @throws Exception
     */
    public String getSign(String appId, long timestamp, String nonceStr, String pack) throws Exception{
        String message = buildMessage(appId, timestamp, nonceStr, pack);
        String paySign= sign(message.getBytes("utf-8"));
        return paySign;
    }

    private String buildMessage(String appId, long timestamp, String nonceStr, String pack) {
        return appId + "\n"
                + timestamp + "\n"
                + nonceStr + "\n"
                + pack + "\n";
    }
    private String sign(byte[] message) throws Exception{
        Signature sign = Signature.getInstance("SHA256withRSA");
        //这里需要一个PrivateKey类型的参数，就是商户的私钥。
        sign.initSign(this.getPrivateKey(PayUtils.privateKeyPath));
        sign.update(message);
        return Base64.getEncoder().encodeToString(sign.sign());
    }

    public static void main(String[] args) {
        String ss = Md5Utils.hash("appId=wxd678efh567hg6787" +
                "&" +
                "nonceStr=5K8264ILTKCH16CQ2502SI8ZNMTM67VS" +
                "&" +
                "package=prepay_id=wx2017033010242291fcfe0db70013231072" +
                "&" +
                "signType=MD5&timeStamp=1490840662" +
                "&" +
                "key=qazwsxedcrfvtgbyhnujmikolp111111");
        System.out.println(ss.toUpperCase());
    }
}
